
@import 'meld.css';

@define-color insert-bg #227000;
@define-color insert-outline shade(@insert-bg, 1.5);
@define-color insert-text #4E9A06;
@define-color delete-bg @theme_bg_color;
@define-color delete-outline shade(@delete-bg, 1.5);
@define-color delete-text #a40000;
@define-color replace-bg #00458c;
@define-color replace-outline shade(@replace-bg, 1.5);
@define-color replace-text #2d6cfc;
@define-color conflict-bg #7a2a28;
@define-color conflict-outline shade(@conflict-bg, 1.5);
@define-color conflict-text #ff0000;
@define-color error-bg #fce94f;
@define-color error-outline shade(@error-bg, 1.5);
@define-color error-text #faad3d;
@define-color inline-bg #3465a4;
@define-color unknown-text #aaaaaa;
@define-color current-line-highlight #111100;
@define-color syncpoint-outline #bbbbbb;
@define-color current-chunk-highlight alpha(#ffffff, 0.1);
